/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###PROGRAM###
Name:      Find Files
Author:    Rocklyte systems
Copyright: Rocklyte Systems, 2003.  All rights reserved.
###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/data/config.h>
#include <pandora/input/clipboard.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/tools/dialog.h>
#include <pandora/tools/filedialog.h>
#include <pandora/data/xml.h>
#include <pandora/startup.h>

STRING ProgName      = "Find Files";
STRING ProgAuthor    = "Rocklyte Systems";
STRING ProgDate      = "December 2003";
STRING ProgCopyright = "Copyright Rocklyte Systems (c) 2003.  All rights reserved.";
LONG   ProgDebug = 0;
FLOAT  ProgKernelVersion = 3.3;

#define PROGVERSION ((FLOAT)1.1)

extern struct KernelBase *KernelBase;
extern UBYTE glAppScript[];

static struct StringsBase *StringsBase;
static struct FileSystemBase *FileSystemBase;
static OBJECTPTR StringsMod = NULL, FileSystemMod = NULL;

static OBJECTPTR glContentInput = NULL, glPathInput = NULL, glFilterInput = NULL, glView = NULL;
static OBJECTPTR glTask, glConfig = NULL, glFindFiles = NULL;

/*****************************************************************************
** Prototypes
*/

static ERROR PROGRAM_ActionNotify(OBJECTPTR, struct acActionNotify *);
static ERROR PROGRAM_Custom(OBJECTPTR, struct acCustom *);
static ERROR PROGRAM_Timer(OBJECTPTR, struct acTimer *);

static ERROR MessageHandler(struct Message *);
static ERROR ExtractFile(OBJECTPTR, struct XMLTag *, STRING, LONG, WORD);
static void RefreshView(void);

#define EXF_FILENAME 0x0001
#define EXF_PATH     0x0002

/*****************************************************************************
** Internal: Failure()
*/

ERROR Failure(STRING Message)
{
   if (StringsMod)     Action(AC_Free, StringsMod, NULL);
   if (FileSystemMod)  Action(AC_Free, FileSystemMod, NULL);
   if (glConfig)       Action(AC_Free, glConfig, NULL);
   if (glFilterInput)  ReleaseObject(glFilterInput);
   if (glPathInput)    ReleaseObject(glPathInput);
   if (glContentInput) ReleaseObject(glContentInput);
   if (glView)         ReleaseObject(glView);

   if (Message) {
      print(Message);
      return(ERR_Failed);
   }
   else return(ERR_Okay);
}

/*****************************************************************************
** Main.
*/

ERROR program(void)
{
   LONG i;
   STRING filter = "*", path = "athene:", content = "";
   OBJECTPTR script, file;
   OBJECTID object_id;
   STRING *args;
   OBJECTID win_target;
   APTR *actions;

   SetResource(RES_MESSAGEHANDLER, (LARGE)&MessageHandler);

   glTask = CurrentTask();

   /*** Load required modules ***/

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ERR_GetField);
      }
   }
   else return(ERR_CreateObject);

   if (CreateObject(ID_MODULE, NULL, &FileSystemMod, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(FileSystemMod, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ERR_GetField);
      }
   }
   else return(ERR_CreateObject);

   /*** Intercept some actions that are required for our program ***/

   if (GetField(glTask, FID_Actions, FT_POINTER, &actions) IS ERR_Okay) {
      actions[AC_ActionNotify] = PROGRAM_ActionNotify;
      actions[AC_Custom]       = PROGRAM_Custom;
      actions[AC_Timer]        = PROGRAM_Timer;
   }

   SubscribeTimer(glTask->UniqueID, 250); /* 1/4 second updates */

   /*** Read the saved settings from our last usage ***/

   if (CreateObject(ID_CONFIG, NULL, &glConfig, NULL,
      FID_Location|TSTRING, "user:config/programs/findfiles.cfg",
      TAGEND) IS ERR_Okay) {
      struct mtReadConfig read;

      read.Section = "Variables";
      read.Item    = "Filter";
      if (Action(MT_ReadConfig, glConfig, &read) IS ERR_Okay) filter = read.Data;

      read.Item = "Content";
      if (Action(MT_ReadConfig, glConfig, &read) IS ERR_Okay) content = read.Data;

      read.Item = "Path";
      if (Action(MT_ReadConfig, glConfig, &read) IS ERR_Okay) path = read.Data;
   }

   /*** Process any arguments passed to our program ***/

   if ((GetField(glTask, FID_ArgsList, FT_POINTER, &args) IS ERR_Okay) AND (args)) {
      for (i=0; args[i]; i++) {
         if ((StrCompare(args[i], "-filter", NULL, STR_MATCHLENGTH) IS ERR_Okay) AND (args[i+1])) {
            filter = args[i+1];
         }
         else if ((StrCompare(args[i], "-path", NULL, STR_MATCHLENGTH) IS ERR_Okay) AND (args[i+1])) {
            path = args[i+1];
         }
         else if ((StrCompare(args[i], "-content", NULL, STR_MATCHLENGTH) IS ERR_Okay) AND (args[i+1])) {
            content = args[i+1];
         }
         else if ((StrCompare(args[i], "-target", NULL, STR_MATCHLENGTH) IS ERR_Okay) AND (args[i+1])) {
            FastFindObject(args[i+1], NULL, &win_target, 1, NULL);
         }
      }
   }

   /*** Run our script to create the interface ***/

   if (CreateObject(ID_SCRIPT, NULL, &script, NULL,
      FID_String|TSTRING,   glAppScript,
      FID_Target|TLONG,     win_target,
      TAGEND) IS ERR_Okay) {
      SetFields(script, FID_Path|TUNLISTED,    path,
                        FID_Content|TUNLISTED, content,
                        FID_Filter|TUNLISTED,  filter,
                        TAGEND);
      Action(AC_Activate, script, NULL);
      Action(AC_Free, script, NULL);


      FastFindObject("view", ID_VIEW, &object_id, 1, NULL);
      if (AccessObject(object_id, 0, &glView) != ERR_Okay) return(Failure("Failed to access the file view."));

      FastFindObject("txtContent", ID_INPUT, &object_id, 1, NULL);
      if (AccessObject(object_id, 0, &glContentInput) != ERR_Okay) return(Failure("Failed to access input box."));

      FastFindObject("txtPath", ID_INPUT, &object_id, 1, NULL);
      if (AccessObject(object_id, 0, &glPathInput) != ERR_Okay) return(Failure("Failed to access input box."));

      FastFindObject("txtfilter", ID_INPUT, &object_id, 1, NULL);
      if (AccessObject(object_id, 0, &glFilterInput) != ERR_Okay) return(Failure("Failed to access input box."));
   }
   else return(Failure("Could not run the application script."));

   /*** MAIN LOOP:  Refer to our MessageHandler() routine to see what goes on in here ***/

   ProcessMessages(NULL, MF_WAIT);

   /*** Save program settings for next time ***/

   if (!glConfig) {
      CreateObject(ID_CONFIG, NULL, &glConfig, NULL, TAGEND);
   }

   if (glConfig) {
      STRING str;

      if (GetField(glFilterInput, FID_String, FT_STRING, &str) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, glConfig, "Variables", "Filter", str);
      }

      if (GetField(glContentInput, FID_String, FT_STRING, &str) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, glConfig, "Variables", "Content", str);
      }

      if (GetField(glPathInput, FID_String, FT_STRING, &str) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, glConfig, "Variables", "Path", str);
      }

      if (CreateObject(ID_FILE, NULL, &file, NULL,
         FID_Location|TSTRING, "user:config/programs/findfiles.cfg",
         FID_Flags|TLONG,      FL_NEW|FL_WRITE,
         TAGEND) IS ERR_Okay) {
         struct acSaveToObject save;

         save.DestinationID = file->UniqueID;
         save.ClassID       = NULL;
         Action(AC_SaveToObject, glConfig, &save);

         Action(AC_Free, file, NULL);
      }
   }

   Failure(NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Function: MessageHandler()
**
** This function is called by the system whenever it decides that we have
** messages that need to be processed.
*/

static ERROR MessageHandler(struct Message *Message)
{
   if (Message->Type IS MSGID_QUIT) {
      /* Returning a terminate code will force all further ProcessMessages()
      ** calls to return.
      */
      return(ERR_Terminate);
   }
   else return(ERR_Failed);
}

/*****************************************************************************
** Program: ActionNotify()
*/

ERROR PROGRAM_ActionNotify(OBJECTPTR Self, struct acActionNotify *Args)
{
   if (Args->Error != ERR_Okay) return(ERR_Okay);

   if (Args->ActionID IS AC_Activate) {
      OBJECTPTR dialog;
      struct mtReadConfig read;
      STRING title, newname;
      OBJECTID configid;
      LONG count, i;

      if (GetClassID(Args->ObjectID) IS ID_DIALOG) {
         if (AccessObject(Args->ObjectID, 5000, &dialog) IS ERR_Okay) {
            GetField(dialog, FID_Title, FT_STRING, &title);

            configid = NULL;
            if ((count = TotalChildren(dialog->UniqueID))) {
               struct ChildEntry list[count];

               if (ListChildren(dialog->UniqueID, list, &count) IS ERR_Okay) {
                  for (i=0; i < count; i++) {
                     if (list[i].ClassID IS ID_CONFIG) {
                        configid = list[i].ObjectID;
                        break;
                     }
                  }
               }
            }

            if (StrCompare("rename", title, sizeof("rename")-1, NULL) IS ERR_Okay) {
               if ((GetField(dialog, FID_InputText, FT_STRING, &newname) IS ERR_Okay) AND (newname) AND (newname[0])) {
                  UBYTE buffer[500];
                  struct Dialog *confirm_dialog;
                  OBJECTPTR config, file;
                  LONG response;

                  if (AccessObject(configid, 5000, &config) IS ERR_Okay) {
                     read.Section = "RENAME";
                     read.Item    = "Location";
                     if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
                        /* Build the name of the new file and check if it
                        ** exists.
                        */

                        StrCopy(read.Data, buffer, sizeof(buffer));
                        for (i=0; buffer[i]; i++);
                        while ((i > 0) AND (buffer[i-1] != '/') AND (buffer[i-1] != '\\') AND (buffer[i-1] != ':')) i--;
                        StrCopy(newname, buffer+i, sizeof(buffer)-i);

                        response = RSF_YES;
                        if (AnalyseLocation(buffer, NULL) IS ERR_Okay) {
                           /* The destination exists - confirm overwrite by creating
                           ** a confirmation dialog.
                           */

                           if (CreateObject(ID_DIALOG, NULL, &confirm_dialog, NULL,
                              FID_Image|TSTRING,   "icons:tools/warning(48)",
                              FID_Buttons|TSTRING, "YESALL:Yes; CANCEL:No",
                              FID_Title|TSTRING,   "Confirmation Required",
                              FID_String|TSTRING,  "This operation will overwrite a file that shares the same file name.  Are you sure that you want to proceed?",
                              FID_Static|TLONG,    TRUE,
                              TAGEND) IS ERR_Okay) {

                              acShow(confirm_dialog);

                              /*** Give the user 20 seconds to respond ***/

                              for (i=0; (i < 2000) AND (!confirm_dialog->Response); i++) {
                                 WaitTime(0, 10000);
                                 ProcessMessages(NULL, NULL);
                              }

                              response = confirm_dialog->Response;

                              acFree(confirm_dialog);
                           }
                        }

                        if (response & RSF_POSITIVE) {
                           if (CreateObject(ID_FILE, NULL, &file, NULL,
                              FID_Location|TSTRING, read.Data,
                              FID_Flags|TLONG,      FL_READ,
                              TAGEND) IS ERR_Okay) {
                              ERROR error;
                              OBJECTPTR error_dialog;

                              error = ActionTags(AC_Rename, file, newname);
                              acFree(file);

                              if (error != ERR_Okay) {
                                 if (CreateObject(ID_DIALOG, NULL, &error_dialog, NULL,
                                    FID_Image|TSTRING,   "icons:items/warning(48)",
                                    FID_Buttons|TSTRING, "NONE:Okay",
                                    FID_Title|TSTRING,   "File Rename Failure",
                                    FID_String|TSTRING,  "The file could not be renamed due to an error.",
                                    TAGEND) IS ERR_Okay) {

                                    acShow(error_dialog);
                                 }
                              }

                              RefreshView();
                           }
                        }
                     }

                     ReleaseObject(config);
                  }
               }
            }
            else if (StrCompare("delete", title, sizeof("delete")-1, NULL) IS ERR_Okay) {
               UBYTE buffer[500];
               struct ConfigEntry *entries;
               OBJECTPTR config, error_dialog;
               STRING errorfile;
               LONG amtentries;
               WORD errorcount;

               if (AccessObject(configid, 5000, &config) IS ERR_Okay) {
                  if (GetFields(config, FID_Entries|TPTR,     &entries,
                                        FID_AmtEntries|TLONG, &amtentries,
                                        TAGEND) IS ERR_Okay) {

                     errorcount = 0;
                     errorfile  = NULL;
                     for (i=0; i < amtentries; i++) {
                        if (DeleteFile(entries[i].Data) != ERR_Okay) {
                           errorfile = entries[i].Data;
                           errorcount++;
                        }
                     }

                     if (errorcount > 0) {
                        DPrintF("FieldNotify:","Errors detected during delete process (%d files)", errorcount);

                        if (errorcount IS 1) StrFormat(buffer, sizeof(buffer), "The following file could not be deleted due to an error:\n\n%s", errorfile);
                        else StrFormat(buffer, sizeof(buffer), "%d files could not be deleted due to errors.", errorcount);

                        if (CreateObject(ID_DIALOG, NULL, &error_dialog, NULL,
                              FID_Image|TSTRING,   "icons:items/warning(48)",
                              FID_Buttons|TSTRING, "NONE:Okay",
                              FID_Title|TSTRING,   "File Deletion Failures",
                              FID_String|TSTRING,  buffer,
                              TAGEND) IS ERR_Okay) {

                           acShow(error_dialog);
                        }
                     }

                     RefreshView();
                  }

                  ReleaseObject(config);
               }
            }

            ReleaseObject(dialog);
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Program: Custom
*/

static ERROR PROGRAM_Custom(OBJECTPTR Self, struct acCustom *Args)
{
   if ((Args) AND (Args->String)) {
      if (StrCompare("FindFiles", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         OBJECTID stop_id, start_id, clear_id;
         STRING str;

         DPrintF("FindFiles:","Initiating file search...");

         if (FastFindObject("btnstop", NULL, &stop_id, 1, NULL) IS ERR_Okay) {
            ActionMsg(AC_Enable, stop_id, NULL);
         }

         if (FastFindObject("btnfindfiles", NULL, &start_id, 1, NULL) IS ERR_Okay) {
            ActionMsg(AC_Disable, start_id, NULL);
         }

         if (FastFindObject("btnclear", NULL, &clear_id, 1, NULL) IS ERR_Okay) {
            ActionMsg(AC_Disable, clear_id, NULL);
         }

         if (CreateObject(ID_FINDFILES, NULL, &glFindFiles, NULL,
            FID_Output|TLONG,  glView->UniqueID,
            FID_Flags|TSTRING, "forcedir|xml",
            TAGEND) IS ERR_Okay) {

            if (GetField(glFilterInput, FID_String, FT_STRING, &str) IS ERR_Okay) {
               SetField(glFindFiles, FID_Filter, FT_STRING, str);
            }

            if (GetField(glContentInput, FID_String, FT_STRING, &str) IS ERR_Okay) {
               SetField(glFindFiles, FID_FileContent, FT_STRING, str);
            }

            if (GetField(glPathInput, FID_String, FT_STRING, &str) IS ERR_Okay) {
               SetField(glFindFiles, FID_Location, FT_STRING, str);
            }

            acActivate(glFindFiles);

            acFree(glFindFiles);
            glFindFiles = NULL;
         }

         if (start_id) ActionMsg(AC_Enable, start_id, NULL);
         if (clear_id) ActionMsg(AC_Enable, clear_id, NULL);
         if (stop_id) ActionMsg(AC_Disable, stop_id, NULL);
      }
      else if (StrCompare("Reset", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         SetField(glContentInput, FID_String, FT_STRING, "");
         SetField(glPathInput, FID_String, FT_STRING, "athene:");
         SetField(glFilterInput, FID_String, FT_STRING, "*");

         Action(AC_Clear, glView, NULL);
      }
      else if (StrCompare("Stop", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         if (glFindFiles) {
            DPrintF("FindFiles:","Stopping the find files process.");
            Action(AC_Deactivate, glFindFiles, NULL);
         }
      }
      else if ((StrCompare("CutFiles", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay) OR
               (StrCompare("CopyFiles", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay)) {
         struct XMLTag **taglist;
         OBJECTID clipid;
         OBJECTPTR clip, xml;
         LONG i, flags, *tags;
         UBYTE buffer[400];

         DPrintF("~CopyFiles()","");

         if (FastFindObject("SystemClipboard", ID_CLIPBOARD, &clipid, 1, NULL) IS ERR_Okay) {
            if (AccessObject(clipid, 5000, &clip) IS ERR_Okay) {
               if (GetFields(glView, FID_SelectedTags|TPTR, &tags,
                                     FID_XML|TPTR,          &xml,
                                     FID_Tags|TPTR,         &taglist,
                                     TAGEND) IS ERR_Okay) {

                  /*** Clear any existing file clips ***/

                  ActionTags(MT_DeleteClip, clip, CLIP_FILE);

                  /*** Add all selected files to the clipboard ***/

                  for (i=0; tags[i] != -1; i++) {
                     if (ExtractFile(xml, taglist[tags[i]], buffer, sizeof(buffer), EXF_PATH|EXF_FILENAME) IS ERR_Okay) {
                        flags = CLF_EXTEND;
                        if (StrCompare("cutfiles", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay) flags |= CLF_DELETE;

                        ActionTags(MT_AddClipboardFiles, clip, CLIP_FILE, buffer, flags);
                     }
                  }
               }

               ReleaseObject(clip);
            }
            else DPrintF("@CopyFiles:","Failed to access clipboard.");
         }

         StepBack();
      }
      else if (StrCompare("RenameFile", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         UBYTE buffer[500];
         OBJECTPTR xml, dialog, config;
         struct XMLTag **taglist;
         LONG tag;

         if (GetFields(glView, FID_XML|TPTR,          &xml,
                               FID_SelectedTag|TLONG, &tag,
                               FID_Tags|TPTR,         &taglist,
                               TAGEND) != ERR_Okay) return(ERR_Failed);

         if (ExtractFile(xml, taglist[tag], buffer, sizeof(buffer), EXF_FILENAME) != ERR_Okay) return(ERR_Failed);

         if (CreateObject(ID_DIALOG, NULL, &dialog, NULL,
            FID_Image|TSTRING,     "icons:tools/edit(48)",
            FID_Buttons|TSTRING,   "okay;cancel",
            FID_Title|TSTRING,     "Rename",
            FID_Flags|TLONG,       DF_INPUT|DF_INPUTREQUIRED,
            FID_String|TSTRING,    "Please enter the new name for the selected file or directory.",
            FID_InputText|TSTRING, buffer,
            TAGEND) IS ERR_Okay) {

            /* Create a Config object that will store the location of
            ** the file that we are renaming.
            */

            if (CreateObject(ID_CONFIG, NULL, &config, NULL,
               FID_Owner|TLONG, dialog->UniqueID,
               TAGEND) IS ERR_Okay) {

               if (ExtractFile(xml, taglist[tag], buffer, sizeof(buffer), EXF_PATH|EXF_FILENAME) IS ERR_Okay) {
                  ActionTags(MT_WriteConfig, config, "RENAME", "Location", buffer);
               }
            }

            Action(AC_Show, dialog, NULL);

            SubscribeAction(dialog, AC_Activate, glTask);

            return(ERR_Okay);
         }
         else return(ERR_CreateObject);
      }
      else if (StrCompare("DeleteFiles", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         OBJECTPTR dialog, xml, config;
         struct XMLTag **taglist;
         WORD i;
         LONG *tags;
         UBYTE buffer[300], intstr[20];

         if (GetFields(glView, FID_SelectedTags|TPTR, &tags,
                               FID_XML|TPTR,          &xml,
                               FID_Tags|TPTR,         &taglist,
                               TAGEND) IS ERR_Okay) {

            for (i=0; tags[i] != -1; i++);

            if (i > 1) StrFormat(buffer, sizeof(buffer), "Are you sure that you wish to delete the %d selected items?", i);
            else {
                i = StrCopy("Are you sure that you wish to delete the '", buffer, sizeof(buffer));

                ExtractFile(xml, taglist[tags[0]], buffer+i, sizeof(buffer)-i, EXF_PATH|EXF_FILENAME);
                while (buffer[i]) i++;

                if (buffer[i-1] IS '/') StrCopy("' folder?", buffer+i-1, sizeof(buffer)-i);
                else StrCopy("' file?", buffer+i, sizeof(buffer)-i);
            }

            if (CreateObject(ID_DIALOG, NULL, &dialog, NULL,
               FID_Image|TSTRING,   "icons:tools/eraser(48)",
               FID_Buttons|TSTRING, "YESALL:Yes; CANCEL:No",
               FID_Title|TSTRING,   "Delete Items",
               FID_String|TSTRING,  buffer,
               TAGEND) IS ERR_Okay) {

               /* Create a Config object that will store all of the
               ** files that we are going to delete.  When the user
               ** responds to the dialog box positively, we'll use the
               ** config object's content to determine what we're going
               ** to delete (see the FieldNotify support action).
               */

               if (CreateObject(ID_CONFIG, NULL, &config, NULL,
                  FID_Owner|TLONG, dialog->UniqueID,
                  TAGEND) IS ERR_Okay) {

                  for (i=0; tags[i] != -1; i++) {
                     if (ExtractFile(xml, taglist[tags[i]], buffer, sizeof(buffer), EXF_PATH|EXF_FILENAME) IS ERR_Okay) {
                        IntToStr(i, intstr, sizeof(intstr));
                        ActionTags(MT_WriteConfig, config, "DELETEFILES", intstr, buffer);
                     }
                  }
               }

               SubscribeAction(dialog, AC_Activate, glTask);

               Action(AC_Show, dialog, NULL);
            }
         }
      }
      else if (StrCompare("EditFiles", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         OBJECTPTR run, xml;
         struct XMLTag **taglist;
         LONG *tags;
         UBYTE buffer[300];
         WORD i;

         if ((GetFields(glView, FID_SelectedTags|TPTR, &tags,
                                FID_XML|TPTR,          &xml,
                                FID_Tags|TPTR,         &taglist,
                                TAGEND) IS ERR_Okay) AND (tags)) {
            for (i=0; tags[i] != -1; i++) {
               if (ExtractFile(xml, taglist[tags[i]], buffer, sizeof(buffer), EXF_PATH|EXF_FILENAME) IS ERR_Okay) {
                  if (CreateObject(ID_RUN, NULL, &run, NULL,
                     FID_Mode|TSTRING,     "edit",
                     FID_Location|TSTRING, buffer,
                     TAGEND) IS ERR_Okay) {
                     Action(AC_Activate, run, NULL);
                     Action(AC_Free, run, NULL);
                  }
               }
            }
         }
      }
      else if (StrCompare("ViewFiles", Args->String, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         OBJECTPTR run, xml;
         struct XMLTag **taglist;
         LONG *tags;
         UBYTE buffer[300];
         WORD i;

         if ((GetFields(glView, FID_SelectedTags|TPTR, &tags,
                                FID_XML|TPTR,          &xml,
                                FID_Tags|TPTR,         &taglist,
                                TAGEND) IS ERR_Okay) AND (tags)) {
            for (i=0; tags[i] != -1; i++) {
               if (ExtractFile(xml, taglist[tags[i]], buffer, sizeof(buffer), EXF_PATH|EXF_FILENAME) IS ERR_Okay) {
                  if (CreateObject(ID_RUN, NULL, &run, NULL,
                     FID_Mode|TSTRING,     "view",
                     FID_Location|TSTRING, buffer,
                     TAGEND) IS ERR_Okay) {
                     Action(AC_Activate, run, NULL);
                     Action(AC_Free, run, NULL);
                  }
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Program: Timer()
*/

static ERROR PROGRAM_Timer(OBJECTPTR Self, struct acTimer *Args)
{
   RefreshView();
   return(ERR_Okay);
}

/*****************************************************************************
** Function: ExtractFile()
*/

static ERROR ExtractFile(OBJECTPTR XML, struct XMLTag *Tag, STRING Buffer, LONG BufferSize, WORD Flags)
{
   struct XMLTag *child;
   WORD j;

   Buffer[0] = 0;

   if ((child = Tag->Child)) {
      if (Flags & EXF_PATH) {
         while (child) {
            if (StrCompare("folder", child->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               ActionTags(MT_GetXMLContent, XML, child->Index, Buffer, BufferSize);
               break;
            }
            child = child->Next;
         }

         if (!Buffer[0]) return(ERR_Failed);
      }

      if (Flags & EXF_FILENAME) {
         /*** Extend path with the file name ***/

         j = StrLength(Buffer);
         ActionTags(MT_GetXMLContent, XML, Tag->Index, Buffer+j, BufferSize-j);
         return(ERR_Okay);
      }
      else return(ERR_Failed);
   }
   else return(ERR_Failed);
}

/*****************************************************************************
** Function: RefreshView()
*/

static void RefreshView(void)
{
   static LONG index = 0;
   OBJECTPTR xml;
   struct XMLTag **taglist, *tag;
   UBYTE file[500];
   BYTE refresh;
   WORD loop_check;
   LONG i;

   refresh = FALSE;
   loop_check = 0;
restart:
   loop_check++;
   if (loop_check < 1000) {
      if ((GetFields(glView, FID_Tags|TPTR, &taglist,
                             FID_XML|TPTR,  &xml,
                             TAGEND) IS ERR_Okay) AND (taglist) AND (xml)) {

         for (i=0, tag=taglist[0]; (i < index) AND (tag); i++) tag = tag->Next;

         for (i=0; (i < 10) AND (tag); i++) {
            if (ExtractFile(xml, tag, file, sizeof(file), EXF_PATH|EXF_FILENAME) IS ERR_Okay) {
               if (AnalyseLocation(file, NULL) != ERR_Okay) {
                  /* The file does not exist, so remove it from the view and restart
                  ** our scanning process.
                  */

                  DPrintF("FindFiles:","Removing \"%s\"", file);

                  if (ActionTags(MT_RemoveXMLTag, xml, tag->Index, 1) IS ERR_Okay) {
                     refresh = TRUE;
                     goto restart;
                  }
               }
            }

            tag = tag->Next;
            index++;
         }

         if (!tag) index = 0; /* Reset the counter for the next round if we have checked all tags */
      }
   }
   else acClear(glView); /* Clear the entire view due to excessive looping */

   if (refresh IS TRUE) {
      acRefresh(glView);
   }
}

